{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Last time we saw treeNode we created hardcoded trees like below:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [],
   "source": [
    "class treeNode:\n",
    "    def __init__(self, val, left=None, right=None):\n",
    "        self.val = val\n",
    "        self.left = left\n",
    "        self.right = right\n",
    "\n",
    "tn1 = treeNode(1)\n",
    "\n",
    "tn2 = treeNode(2)\n",
    "\n",
    "tn3 = treeNode(3)\n",
    "\n",
    "tn4 = treeNode(4)\n",
    "\n",
    "tn5 = treeNode(5, tn1, tn2)\n",
    "\n",
    "tn6 = treeNode(6, tn3, tn4)\n",
    "\n",
    "tn7 = treeNode(7, tn5, tn6)\n",
    "\n",
    "# Much like linked list, our treeNode can be used to make a binaryTree object\n",
    "# Right now we dont have any real use for this though...\n",
    "# Still you should understand that storing the root node of a tree is equivalent\n",
    "# to storing the full tree (like head stores the full linked list)\n",
    "class binaryTree:\n",
    "    def __init__(self):\n",
    "        self.root = None"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Today we will use randomly generated trees. Also provided are two helpful support functions `height()` (covered in class previously) and `printTree()` (not covered)."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {},
   "outputs": [],
   "source": [
    "import random\n",
    "\n",
    "def build_random_tree(size, seed=None):\n",
    "    random.seed(seed)\n",
    "    keys = list(range(size))\n",
    "    random.shuffle(keys)\n",
    "\n",
    "    root = random_tree_helper(keys)\n",
    "    return root\n",
    "\n",
    "def random_tree_helper(keyList):\n",
    "    # Base Case -- keyList has 0 or 1 nodes left\n",
    "    if len(keyList) == 0:\n",
    "        return None\n",
    "    if len(keyList) == 1:\n",
    "        return treeNode(keyList[0])\n",
    "    \n",
    "    # Reduction Step -- Remove front of keyList and make it a new node\n",
    "    node = treeNode(keyList.pop(0))\n",
    "    \n",
    "    # Combining Step -- Split the remaining keyList in a random partition\n",
    "    partition = random.randint(0, len(keyList))\n",
    "    leftList = keyList[:partition]\n",
    "    rightList = keyList[partition:]\n",
    "\n",
    "    node.left = random_tree_helper(leftList)\n",
    "    node.right = random_tree_helper(rightList)\n",
    "    \n",
    "    return node\n",
    "\n",
    "# INPUT:\n",
    "# A treeNode object storing the root of the tree being measured (node)\n",
    "# OUTPUT:\n",
    "# An integer storing the height of the tree defined by root\n",
    "# Hint: A recursive solution is easiest (as seen in class)!\n",
    "def height(node):\n",
    "    if node == None:\n",
    "        return -1\n",
    "\n",
    "    return 1 + max(height(node.left), height(node.right))\n",
    "\n",
    "# INPUT:\n",
    "# A treeNode object storing the root of the tree being printed (root)\n",
    "# OUTPUT:\n",
    "# Nothing\n",
    "# Prints to terminal an ASCII interpretation of the graph\n",
    "# IMPORTANT: Requires a functional implementation of height!\n",
    "def printTree(root):\n",
    "    if root == None:\n",
    "        print(\"(empty)\\n\")\n",
    "        return\n",
    "\n",
    "    root_height = height(root)\n",
    "    print_matrix_width = (4 << root_height) - 3\n",
    "    print_matrix_height = 2 * root_height + 1\n",
    "\n",
    "    output = [None]*print_matrix_height\n",
    "    for i in range(len(output)):\n",
    "        output[i] = [' '] * (print_matrix_width+4)\n",
    "\n",
    "    pt_helper(root, output, 0, 0, print_matrix_width)\n",
    "\n",
    "    for row in output:\n",
    "        print(\"\".join(row))\n",
    "\n",
    "\n",
    "def pt_helper(root, output, left, top, curr_width):\n",
    "    val = [char for char in str(root.val)]\n",
    "    curr_width = int(curr_width)\n",
    "\n",
    "    left_start_shift = int(1 - (len(val)-1) / 2)\n",
    "    i = 0\n",
    "    while (i < len(val) and left + curr_width/2 + i < len(output[top])):\n",
    "        output[top][int(left + curr_width/2 +\n",
    "                        left_start_shift + i)] = val[i]\n",
    "        i += 1\n",
    "\n",
    "    branchOffset = (curr_width+3) >> 3\n",
    "    branchOffset = int(branchOffset)\n",
    "\n",
    "    center = int(left + curr_width/2)\n",
    "    leftcenter = int(left + (curr_width/2 - 1)/2)\n",
    "    rightcenter = int(left + curr_width/2 + 2 + (curr_width/2 - 1)/2)\n",
    "\n",
    "    if root.left != None:\n",
    "        branch_pos = center - branchOffset + 1\n",
    "\n",
    "        for pos in range(int(center+left_start_shift - 2), branch_pos, -1):\n",
    "            output[top][pos] = '_'\n",
    "\n",
    "        output[top+1][branch_pos] = '/'\n",
    "\n",
    "        for pos in range(branch_pos-1, leftcenter + 2, -1):\n",
    "            output[top+1][pos] = '_'\n",
    "\n",
    "        pt_helper(root.left, output, left, top+2, curr_width/2 - 1)\n",
    "\n",
    "    if root.right != None:\n",
    "        branch_pos = center + branchOffset + 1\n",
    "\n",
    "        for pos in range(center+left_start_shift + len(val) + 1, branch_pos, 1):\n",
    "            output[top][pos] = '_'\n",
    "\n",
    "        output[top+1][branch_pos] = '\\\\'\n",
    "\n",
    "        for pos in range(branch_pos+1, rightcenter, 1):\n",
    "            output[top+1][pos] = '_'\n",
    "\n",
    "        pt_helper(root.right, output, left +\n",
    "                  curr_width/2 + 2, top+2, curr_width/2 - 1)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "   1     \n",
      "  / \\    \n",
      " 2    0  \n",
      "*****\n",
      "       1         \n",
      "         \\_      \n",
      "            2    \n",
      "           /     \n",
      "          0      \n"
     ]
    }
   ],
   "source": [
    "printTree(build_random_tree(3, 1))\n",
    "print(\"*****\")\n",
    "printTree(build_random_tree(3, 1001))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Lets code up binary tree insert according to the class brainstorm specifications!"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "            __ 2 __              \n",
      "         __/       \\___          \n",
      "       3                6        \n",
      "     /   \\_                      \n",
      "   1        5                    \n",
      "  /          \\                   \n",
      " 4            0                  \n",
      "*****\n",
      "                        ______ 2 ______                          \n",
      "                 ______/               \\_______                  \n",
      "            __ 3 __                             6                \n",
      "         __/       \\___                                          \n",
      "      12                5                                        \n",
      "     /                    \\                                      \n",
      "   1                        0                                    \n",
      "  /                                                              \n",
      " 4                                                               \n"
     ]
    }
   ],
   "source": [
    "# INPUT:\n",
    "# node is parent\n",
    "# pos is 'left' or 'right'\n",
    "# val is val\n",
    "def bt_insert(node, pos, val):\n",
    "    # We must look to see if a child exists\n",
    "    # If it exists, we need to save that branch\n",
    "    if pos == 'left':\n",
    "        # Look at node.left\n",
    "        if node.left == None: # This position is open!\n",
    "            node.left = treeNode(val)\n",
    "        else:\n",
    "            # Keep track of old node\n",
    "            tmp = node.left\n",
    "            # Assign new node to position\n",
    "            node.left = treeNode(val)\n",
    "            # Add old branch as child of our new node\n",
    "            node.left.left = tmp\n",
    "        \n",
    "\n",
    "    \n",
    "\n",
    "root1 = build_random_tree(7, 2)\n",
    "printTree(root1)\n",
    "print(\"*****\")\n",
    "# We are GIVEN the node we want\n",
    "node6 = root1.left\n",
    "bt_insert(node6, 'left', 12)\n",
    "printTree(root1)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "            __ 3 __              \n",
      "         __/       \\___          \n",
      "       5                2        \n",
      "     /                /   \\      \n",
      "   1                4       0    \n",
      "    \\                            \n",
      "      6                          \n",
      "****\n",
      "3\n",
      "5\n",
      "1\n",
      "6\n",
      "2\n",
      "4\n",
      "0\n"
     ]
    }
   ],
   "source": [
    "\n",
    "# In-class Exercise: Pre-order Traversal\n",
    "def preorderTraversal(node):\n",
    "    # I'm missing a base case!\n",
    "    if node: #if node is None, dont do anything!\n",
    "    # I look at myself first\n",
    "        print(node.val)\n",
    "        preorderTraversal(node.left)\n",
    "        preorderTraversal(node.right)\n",
    "\n",
    "\n",
    "random.seed(None)\n",
    "root1 = build_random_tree(7)\n",
    "printTree(root1)\n",
    "print(\"****\")\n",
    "preorderTraversal(root1)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "**Binary Tree Search** In order to efficiently search our trees, we have to bring back a rather useful pair of data structures!"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Node:\n",
    "    def __init__(self, value, next = None):\n",
    "         self.val = value\n",
    "         self.next = next\n",
    "\n",
    "class stack: \n",
    "    def __init__(self):\n",
    "        self.head = None\n",
    "        self.length = 0\n",
    "    \n",
    "    def push(self, value):\n",
    "        self.length += 1\n",
    "        newNode = Node(value)\n",
    "        newNode.next = self.head\n",
    "        self.head = newNode\n",
    "    \n",
    "    def __len__(self):\n",
    "        return self.length    \n",
    "    \n",
    "    def top(self):\n",
    "        if self.length > 0:\n",
    "            return self.head.val\n",
    "        return None\n",
    "    \n",
    "    def pop(self):\n",
    "        if (self.length > 0):\n",
    "            self.length -= 1\n",
    "            popped = self.head\n",
    "            self.head = self.head.next\n",
    "            return popped.val\n",
    "        return None\n",
    "    \n",
    "    # Added to make your print easier\n",
    "    def __str__(self):\n",
    "        if self.length == 0:\n",
    "            return \"[]\"\n",
    "        \n",
    "        traversalNode = self.head\n",
    "        out = \"[\"\n",
    "        while (traversalNode != None):\n",
    "            out += str(traversalNode.val) + \", \"\n",
    "            traversalNode = traversalNode.next\n",
    "        out = out[:-2]\n",
    "        out += \"]\"\n",
    "        return out\n",
    "\n",
    "    def empty(self):\n",
    "        return self.length == 0\n",
    "\n",
    "class queue:\n",
    "    def __init__(self):\n",
    "        self.length = 0\n",
    "        self.head = None\n",
    "        self.tail = None\n",
    "    \n",
    "    def enqueue(self, value):\n",
    "        self.length += 1\n",
    "        item = Node(value)\n",
    "        if self.head == None:\n",
    "            self.head = item\n",
    "            self.tail = item\n",
    "        else:\n",
    "            self.tail.next = item\n",
    "            self.tail = self.tail.next\n",
    "            \n",
    "    def dequeue(self):\n",
    "        if self.length > 0:\n",
    "            self.length -= 1\n",
    "            item = self.head\n",
    "            self.head = item.next\n",
    "\n",
    "            if self.head == None:\n",
    "                self.tail = None\n",
    "            return item.val\n",
    "\n",
    "        return None\n",
    "\n",
    "    def front(self):\n",
    "        if self.length > 0:\n",
    "            return self.head.val\n",
    "        return None\n",
    "    \n",
    "    def __str__(self):\n",
    "        if self.length == 0:\n",
    "            return \"[]\"\n",
    "        \n",
    "        traversalNode = self.head\n",
    "        out = \"[\"\n",
    "        while (traversalNode != None):\n",
    "            out += str(traversalNode.val) + \", \"\n",
    "            traversalNode = traversalNode.next\n",
    "        out = out[:-2]\n",
    "        out += \"]\"\n",
    "        return out\n",
    "\n",
    "    def __len__(self):\n",
    "        return self.length\n",
    "\n",
    "    def empty(self):\n",
    "        return self.length == 0"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 16,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "            __ 5 __              \n",
      "         __/       \\___          \n",
      "       3                6        \n",
      "     /   \\_           /          \n",
      "   4        1       7            \n",
      "             \\       \\           \n",
      "              2       0          \n",
      "***\n",
      "5\n",
      "6\n",
      "7\n",
      "0\n",
      "3\n",
      "1\n",
      "True\n",
      "***\n",
      "5\n",
      "6\n",
      "7\n",
      "0\n",
      "3\n",
      "1\n",
      "2\n",
      "4\n",
      "None\n"
     ]
    }
   ],
   "source": [
    "# Oops I accidentally did traversal twice\n",
    "# How do I turn this into a search problem?\n",
    "\n",
    "def depthfirstSearch(root, val):\n",
    "    s = stack()\n",
    "    #Initialize our stack\n",
    "    s.push(root)\n",
    "    while len(s)!=0:\n",
    "        n = s.pop() \n",
    "        # Print my value\n",
    "        print(n.val)\n",
    "        # Now I want to not print value\n",
    "        if n.val == val:\n",
    "            return True\n",
    "        # 'Recurse' but iteratively!\n",
    "        # What are the children of n\n",
    "        # Only add to stack if value is not None\n",
    "        if n.left != None:\n",
    "            s.push(n.left) # left child (CAN BE NONE)!\n",
    "        if n.right != None:\n",
    "            s.push(n.right) # right child!\n",
    "    return False\n",
    "\n",
    "root = build_random_tree(8, 2)\n",
    "printTree(root)\n",
    "print(\"***\")\n",
    "print(depthfirstSearch(root, 1))\n",
    "print(\"***\")\n",
    "print(depthfirstSearch(root, 10))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 13,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "            __ 5 __              \n",
      "         __/       \\___          \n",
      "       3                6        \n",
      "     /   \\_           /          \n",
      "   4        1       7            \n",
      "             \\       \\           \n",
      "              2       0          \n",
      "***\n",
      "None\n",
      "***\n"
     ]
    }
   ],
   "source": [
    "def breadthfirstSearch(root, val):\n",
    "    pass\n",
    "\n",
    "root = build_random_tree(8, 2)\n",
    "printTree(root)\n",
    "print(\"***\")\n",
    "print(breadthfirstSearch(root, 1))\n",
    "print(\"***\")\n",
    "breadthfirstSearch(root, 6)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.11.0"
  },
  "vscode": {
   "interpreter": {
    "hash": "1a1af0ee75eeea9e2e1ee996c87e7a2b11a0bebd85af04bb136d915cefc0abce"
   }
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
